@@ -8,14 +8,18 @@ import itertools
 import json
 import subprocess
 import sys
+import time
 import urllib.parse
-from typing import List, Optional, TypeVar
+from functools import partial
+from typing import BinaryIO, Callable, List, Optional, TypeVar
 
 import colorama
 import gitlab
+import requests.exceptions
+from gitlab.v4.objects import ProjectJob
 
 
-__version__ = '0.6.2'
+__version__ = '0.7.1'
 __author__ = "Marius Gedminas <marius@gedmin.as>"
 
 
@@ -27,7 +31,7 @@ def fatal(msg: str) -> None:
 
 
 def warn(msg: str) -> None:
-    print(msg, file=sys.stderr)
+    print(msg, file=sys.stderr, flush=True)
 
 
 def info(msg: str) -> None:
@@ -95,6 +99,38 @@ def fmt_size(size: Optional[float]) -> str:
     return f'{size:.1f}'.rstrip('0').rstrip('.') + f' {unit}'
 
 
+def tail(s: bytes, n: Optional[int] = None) -> bytes:
+    if not n:
+        return s
+    return b''.join(s.splitlines(True)[-n:])
+
+
+def follow(
+    job: ProjectJob, buffer: Optional[BinaryIO] = None, interval: float = 1.0,
+    tail: Optional[Callable[[bytes], bytes]] = None
+) -> None:
+    if buffer is None:
+        buffer = sys.stdout.buffer
+    if tail is None:
+        tail = lambda s: s  # noqa: E731
+    trace = job.trace()
+    buffer.write(tail(trace))
+    buffer.flush()
+    while not job.finished_at:
+        time.sleep(interval)
+        job.refresh()
+        new_trace = job.trace()
+        if not new_trace.startswith(trace):
+            # maybe the beginning got truncated?
+            warn("\n----- trace was truncated -----")
+            trace = ""
+        new_data = new_trace[len(trace):]
+        if new_data:
+            buffer.write(new_data)
+            buffer.flush()
+        trace = new_trace
+
+
 def _main() -> None:
     colorama.init()
 
@@ -126,6 +162,13 @@ def _main() -> None:
         "--job", metavar="ID",
         help="show the trace of GitLab CI job with this ID",
     )
+    parser.add_argument(
+        "--running", action="store_true",
+        help=(
+            "show the trace of the currently running GitLab CI job,"
+            " if there is one (if there's more than one, picks the first one)"
+        ),
+    )
     parser.add_argument(
         "-b", "--branch", "--ref", metavar="NAME",
         help=(
@@ -133,6 +176,14 @@ def _main() -> None:
             " (default: the currently checked out branch)"
         ),
     )
+    parser.add_argument(
+        "-t", "--tail", metavar='N', nargs='?', type=int, const=10,
+        help="show the last N lines of the trace log",
+    )
+    parser.add_argument(
+        "-f", "--follow", action="store_true",
+        help="periodically poll and output additional logs as the job runs",
+    )
     parser.add_argument(
         "--print-url", "--print-uri", action="store_true",
         help="print URL to job page on GitLab instead of printing job's log",
@@ -161,6 +212,9 @@ def _main() -> None:
     )
     args = parser.parse_args()
 
+    if args.job and args.running:
+        warn(f"Ignoring --running because --job={args.job} was specified")
+
     if args.job and args.pipeline:
         warn(f"Ignoring pipeline ({args.pipeline})"
              f" because --job={args.job} was specified")
@@ -181,7 +235,7 @@ def _main() -> None:
             info(f"Current branch: {args.branch}")
 
         which = -args.pipeline - 1 if args.pipeline else 0
-        pipelines = project.pipelines.list(ref=args.branch, as_list=False)
+        pipelines = project.pipelines.list(ref=args.branch, iterator=True)
         skipped = sum(1 for p in itertools.islice(pipelines, which))
         pipeline = next(pipelines, None)
         if pipeline is not None:
@@ -233,11 +287,19 @@ def _main() -> None:
             for job in jobs:
                 status = fmt_status(job.status)
                 print(f"   --job={job.id} - {status} - {job.name}")
-            if args.artifacts:
-                warn("Ignoring --artifacts because no job was selected.")
-            if args.print_url:
-                warn("Ignoring --print-url because no job was selected.")
-            sys.exit(0)
+                if args.running and job.status == 'running' and not args.job:
+                    args.job = job.id
+                    job_name = job.name
+            if args.job:
+                info(f"Automatically selected --job={args.job} ({job_name})")
+            else:
+                if args.running:
+                    warn("Ignoring --running because no job was running.")
+                if args.artifacts:
+                    warn("Ignoring --artifacts because no job was selected.")
+                if args.print_url:
+                    warn("Ignoring --print-url because no job was selected.")
+                sys.exit(0)
 
     job = project.jobs.get(args.job)
     if args.verbose:
@@ -249,8 +311,10 @@ def _main() -> None:
         info(json.dumps(job.attributes, indent=2))
     if args.print_url:
         print(f"{project.web_url}/-/jobs/{job.id}")
+    elif args.follow:
+        follow(job, tail=partial(tail, n=args.tail))
     else:
-        sys.stdout.buffer.write(job.trace())
+        sys.stdout.buffer.write(tail(job.trace(), args.tail))
     if args.artifacts:
         if not hasattr(job, 'artifacts_file'):
             warn("Job has no artifacts.")
@@ -265,6 +329,8 @@ def _main() -> None:
 def main() -> None:
     try:
         _main()
+    except requests.exceptions.RequestException as e:
+        sys.exit(str(e))
     except (KeyboardInterrupt, BrokenPipeError):
         # suppress tracebacks from these
         sys.exit(0)
