@@ -1,8 +1,10 @@
 import subprocess
 import sys
 import textwrap
+import time
 
 import pytest
+import requests.exceptions
 
 import gitlab_trace as gt
 
@@ -12,6 +14,11 @@ def mock_gitlab(monkeypatch):
     monkeypatch.setattr(gt, 'gitlab', FakeGitlabModule())
 
 
+@pytest.fixture(autouse=True)
+def mock_time_sleep(monkeypatch):
+    monkeypatch.setattr(time, 'sleep', lambda seconds: None)
+
+
 class FakeGitlabModule:
     __version__ = '0.42.frog-knows'
 
@@ -25,6 +32,8 @@ class FakeGitlabModule:
 
     class Projects:
         def get(self, project_id):
+            if project_id == '404':
+                raise requests.exceptions.HTTPError
             return FakeGitlabModule.Project(project_id)
 
     class Project:
@@ -34,8 +43,8 @@ class FakeGitlabModule:
             self.web_url = f'https://git.example.com/{project_id}'
 
     class ProjectPipelines:
-        def list(self, ref=None, as_list=True):
-            assert not as_list
+        def list(self, ref=None, iterator=False):
+            assert iterator
             if ref == 'empty':
                 return
             else:
@@ -92,12 +101,25 @@ class FakeGitlabModule:
                     'size': 0,
                 }
             self.attributes = {"type": "job", "json_attributes": "here"}
+            self._trace = b'Hello, world!\n'
+            self._refresh = [
+                {},
+                {
+                    '_trace': self._trace + b'Bye!\n',
+                    'finished_at': '2020-09-16T06:16:57.452Z',
+                    'status': 'success',
+                },
+            ]
 
         def artifacts(self, streamed=False, action=None):
             pass
 
+        def refresh(self):
+            if self._refresh:
+                self.__dict__.update(self._refresh.pop(0))
+
         def trace(self):
-            return b'Hello, world!\n'
+            return self._trace
 
 
 def test_fatal():
@@ -184,6 +206,56 @@ def test_fmt_size(size, expected):
     assert gt.fmt_size(size) == expected
 
 
+@pytest.mark.parametrize("s, n, expected", [
+    (b'a\nb\nc\nd\n', None, b'a\nb\nc\nd\n'),
+    (b'a\nb\nc\nd\n', 0, b'a\nb\nc\nd\n'),
+    (b'a\nb\nc\nd\n', 1, b'd\n'),
+    (b'a\nb\nc\nd\n', 2, b'c\nd\n'),
+    (b'a\nb\nc\nd', 1, b'd'),
+])
+def test_tail(s, n, expected):
+    assert gt.tail(s, n) == expected
+
+
+def test_follow_truncation(monkeypatch, capsys):
+    job = FakeGitlabModule.ProjectJob(42, 'job', 'running')
+    job._refresh = [
+        {'_trace': b'Hello, world!\nBye?\n'},
+        {'_trace': b'Bye, cruel world!\n'},
+        {'finished_at': '2020-09-16T06:16:57.452Z'},
+    ]
+    gt.follow(job)
+    stdout, stderr = capsys.readouterr()
+    assert stdout == textwrap.dedent("""\
+        Hello, world!
+        Bye?
+        Bye, cruel world!
+    """)
+    assert stderr == textwrap.dedent("""\
+
+        ----- trace was truncated -----
+    """)
+
+
+def test_follow_truncation_flushes(monkeypatch, capsys):
+    job = FakeGitlabModule.ProjectJob(42, 'job', 'running')
+    job._refresh = [
+        {'_trace': b'Hello, world!\nBye?\n'},
+        {'_trace': b'Bye, cruel world!\n'},
+        {'finished_at': '2020-09-16T06:16:57.452Z'},
+    ]
+    gt.follow(job, buffer=sys.stderr.buffer)
+    stdout, stderr = capsys.readouterr()
+    assert stdout == ''
+    assert stderr == textwrap.dedent("""\
+        Hello, world!
+        Bye?
+
+        ----- trace was truncated -----
+        Bye, cruel world!
+    """)
+
+
 def test_main_help(monkeypatch):
     monkeypatch.setattr(sys, 'argv', ['gitlab-trace', '--help'])
     with pytest.raises(SystemExit):
@@ -191,13 +263,20 @@ def test_main_help(monkeypatch):
 
 
 def test_main_warn_extra_args(monkeypatch, capsys):
-    monkeypatch.setattr(sys, 'argv', ['gitlab-trace', '--job=123', '456'])
+    monkeypatch.setattr(
+        sys, 'argv', ['gitlab-trace', '--running', '--job=123', '456']
+    )
     monkeypatch.setattr(gt, 'determine_project', lambda: None)
     with pytest.raises(SystemExit):
         gt.main()
+    stdout, stderr = capsys.readouterr()
+    assert (
+        'Ignoring --running because --job=123 was specified'
+        in stderr
+    )
     assert (
         'Ignoring pipeline (456) because --job=123 was specified'
-        in capsys.readouterr().err
+        in stderr
     )
 
 
@@ -303,6 +382,27 @@ def test_main_artifacts_no_job_selected(monkeypatch, capsys):
     """)
 
 
+def test_main_no_running_job(monkeypatch, capsys):
+    monkeypatch.setattr(sys, 'argv', ['gitlab-trace', '--running', '-a'])
+    monkeypatch.setattr(gt, 'determine_project', lambda: 'owner/project')
+    monkeypatch.setattr(gt, 'determine_branch', lambda: 'main')
+    with pytest.raises(SystemExit):
+        gt.main()
+    stdout, stderr = capsys.readouterr()
+    assert stderr == textwrap.dedent("""\
+        GitLab project: owner/project
+        Current branch: main
+        https://git.example.com/owner/project/pipelines/1005
+        Ignoring --running because no job was running.
+        Ignoring --artifacts because no job was selected.
+    """)
+    assert stdout == textwrap.dedent("""\
+        Available jobs for pipeline #1005:
+           --job=3201 - success - build
+           --job=3202 - failed - test
+    """)
+
+
 def test_main_pipeline_debug(monkeypatch, capsys):
     monkeypatch.setattr(sys, 'argv', ['gitlab-trace', '--debug'])
     monkeypatch.setattr(gt, 'determine_project', lambda: 'owner/project')
@@ -437,6 +537,27 @@ def test_main_job_by_name_multiple_nth(monkeypatch, capsys):
     """)
 
 
+def test_main_running(monkeypatch, capsys):
+    monkeypatch.setattr(sys, 'argv', ['gitlab-trace', '1009', '--running'])
+    monkeypatch.setattr(gt, 'determine_project', lambda: 'owner/project')
+    monkeypatch.setattr(gt, 'determine_branch', lambda: 'refactoring')
+    with pytest.raises(SystemExit):
+        gt.main()
+    stdout, stderr = capsys.readouterr()
+    assert stderr == textwrap.dedent("""\
+        GitLab project: owner/project
+        Automatically selected --job=3304 (test)
+    """)
+    assert stdout == textwrap.dedent("""\
+        Available jobs for pipeline #1009:
+           --job=3301 - success - build
+           --job=3302 - failed - test
+           --job=3303 - failed - test
+           --job=3304 - running - test
+        Hello, world!
+    """)
+
+
 def test_main_branch_ignored_because_job(monkeypatch, capsys):
     monkeypatch.setattr(sys, 'argv', [
         'gitlab-trace', '--job=3202', '--branch=foo'])
@@ -484,6 +605,24 @@ def test_main_job_verbose(monkeypatch, capsys):
     """)
 
 
+def test_main_job_follow(monkeypatch, capsys):
+    monkeypatch.setattr(
+        sys, "argv", ["gitlab-trace", "--job=3202", "--follow"]
+    )
+    monkeypatch.setattr(gt, 'determine_project', lambda: 'owner/project')
+    monkeypatch.setattr(gt, 'determine_branch', lambda: 'main')
+    with pytest.raises(SystemExit):
+        gt.main()
+    stdout, stderr = capsys.readouterr()
+    assert stderr == textwrap.dedent("""\
+        GitLab project: owner/project
+    """)
+    assert stdout == textwrap.dedent("""\
+        Hello, world!
+        Bye!
+    """)
+
+
 def test_main_job_debug(monkeypatch, capsys):
     monkeypatch.setattr(sys, 'argv', ['gitlab-trace', '--job=3202', '--debug'])
     monkeypatch.setattr(gt, 'determine_project', lambda: 'owner/project')
@@ -563,3 +702,9 @@ def test_main_suppresses_keyboard_interrupt(monkeypatch, capsys):
     monkeypatch.setattr(gt, 'determine_project', raise_keyboard_interrupt)
     with pytest.raises(SystemExit):
         gt.main()
+
+
+def test_main_reports_network_issues_without_tracebacks(monkeypatch, capsys):
+    monkeypatch.setattr(sys, 'argv', ['gitlab-trace', '--project=404'])
+    with pytest.raises(SystemExit):
+        gt.main()
